#include <opencv2/core/core.hpp>
#include <pybind11/pybind11.h>
#include <pybind11/numpy.h>
#include <pybind11/stl.h>
#include "../../ar_sandbox_lib/inc/sandbox.h"


// Convert a type to a readable opencv Type format.
// https://stackoverflow.com/questions/10167534/how-to-find-out-what-type-of-a-mat-object-is-with-mattype-in-opencv
std::string cvtype2str(int type) {
    std::string r;
    unsigned char depth = type & CV_MAT_DEPTH_MASK;
    unsigned char chans = 1 + (type >> CV_CN_SHIFT);
    switch ( depth ) {
        case CV_8U:  r = "CV_8U"; break;
        case CV_8S:  r = "CV_8S"; break;
        case CV_16U: r = "CV_16U"; break;
        case CV_16S: r = "CV_16S"; break;
        case CV_32S: r = "CV_32S"; break;
        case CV_32F: r = "CV_32F"; break;
        case CV_64F: r = "CV_64F"; break;
        default:     r = "CV_User"; break;
    }
    r += "C";
    r += (chans+'0');
    return r;
}

// Print some info on the provided matrix.
void printMatrixInfo(cv::Mat &matrix)
{
    std::cout << "MATRIX :" << std::endl;
    std::cout << matrix << std::endl;
    std::cout << "Channels: ";
    std::cout << matrix.channels() << std::endl;
    std::cout << "Rows: ";
    std::cout << matrix.rows << std::endl;
    std::cout << "Cols: ";
    std::cout << matrix.cols << std::endl;
    std::cout << "Depth: ";
    std::cout << cvtype2str(matrix.type()) << std::endl;
}

// Remove this
int sz_cols = 5;
int sz_rows = 6;

// Remove this
void populate3Channels(cv::Mat &mat) {
    if (mat.channels() == 3) {
        for (int c = 0; c < mat.channels(); c++) {
            for (int i = 0; i < mat.rows; i++) {
                for (int j = 0;j < mat.cols; j++) {
                    mat.at<cv::Vec3b>(i, j) = cv::Vec3b(i * mat.cols + j, i * mat.cols + j, i * mat.cols + j);
                }
            }
        }
    }
}

// Remove this
void populate1Channel(cv::Mat &mat) {
    for (int i = 0; i < mat.rows; i++) {
        for (int j = 0;j < mat.cols; j++) {
            mat.at<float>(i, j) = i * mat.cols + j;
        }
    }
}
// Give dummy matrixes. Should be deleted.
cv::Mat_<cv::Vec3b> giveMatrix() {
    cv::Mat_<cv::Vec3b> matrix = cv::Mat_<cv::Vec3b>::zeros(sz_cols, sz_rows);
    populate3Channels(matrix);
    // printMatrixInfo(matrix);
    cv::Rect rec(1, 1, 3, 3);
    cv::Mat crop = matrix(rec);
    cv::Mat outmat;
    crop.copyTo(outmat);
    // printMatrixInfo(outmat);
    return outmat;
}

cv::Mat_<float> giveFloatMatrix() {
    cv::Mat_<float> matrix = cv::Mat_<float>::zeros(sz_cols, sz_rows);
    populate1Channel(matrix);
    // printMatrixInfo(matrix);
    cv::Rect rec(1, 1, 3, 3);
    cv::Mat crop = matrix(rec);
    cv::Mat outmat;
    crop.copyTo(outmat);
    // printMatrixInfo(outmat);
    return outmat;
}

// Dummy function to check if parameters are valid, for testing purposes.
// This should obviously be deleted.
int adjProj(Sandbox &self, cv::Mat_<cv::Vec3b> frame, cv::Mat_<float> depth) {
    // printMatrixInfo(frame);
    // printMatrixInfo(depth);
    return 0;
}


// Create a cv::Mat<Vec3b> from a numpy array.
// Used as a constructor for cvMat3b
cv::Mat_<cv::Vec3b> npBufferToCvMat3b(pybind11::array_t<unsigned char>& arr) {
    pybind11::buffer_info buf = arr.request();
    cv::Mat_<cv::Vec3b> m;
    cv::Mat mat(buf.shape[0], buf.shape[1], CV_8UC3, (unsigned char*)buf.ptr);
    m = mat;
    return m;
}

// Create a cv::Mat<float> from a numpy array.
// Used as a constructor for cvMatf
cv::Mat_<float> npBufferToCvMatF(pybind11::array_t<float>& arr) {
    pybind11::buffer_info buf = arr.request();
    cv::Mat_<float> m;
    cv::Mat mat(buf.shape[0], buf.shape[1], CV_32F, (unsigned char*)buf.ptr);
    // return self;
    m = mat;
    return m;
}

// Create a numpy-readable buffer from a cv::Mat_<float>.
pybind11::buffer_info cvMatFToNpBuffer(cv::Mat_<float>& im) {
    return pybind11::buffer_info(
        // Pointer to buffer
        im.data,
        // Size of one element
        sizeof(float),
        // Python struct-style format descriptor
        pybind11::format_descriptor<float>::format(),
        // Number of dimensions
        2,
        // Buffer dimensions
        { im.rows, im.cols },
        // Strides (in bytes) for each index
        {
            sizeof(float) * im.cols,
            sizeof(float)
        });
}

// Create a numpy-readable buffer from a cv::Mat_<Vec3b>.
pybind11::buffer_info cvMat3bToNpBuffer(cv::Mat_<cv::Vec3b>& im) {
    return pybind11::buffer_info(
        // Pointer to buffer
        im.data,
        // Size of one element
        sizeof(unsigned char),
        // Python struct-style format descriptor
        pybind11::format_descriptor<unsigned char>::format(),
        // Number of dimensions
        3,
        // Buffer dimensions
        { im.rows, im.cols, im.channels() },
        // Strides (in bytes) for each index
        {
            sizeof(unsigned char) * im.channels() * im.cols,
            sizeof(unsigned char) * im.channels(),
            sizeof(unsigned char)
        });
}


PYBIND11_MODULE(sandbox_wrapper, m) {
    m.def("new", giveMatrix);
    m.def("newf", giveFloatMatrix);
    m.def("print_matrix_info",[](pybind11::object mat, int type){
        cv::Mat m;
        switch (type) {
            case CV_8UC3:
                m = mat.cast<cv::Mat_<cv::Vec3b>>();
                break;
            case CV_32FC1:
                m = mat.cast<cv::Mat_<float>>();
                break;
            default:
                break;
        }
        return printMatrixInfo(m);
    });

    // cvMat classes
    pybind11::class_<cv::Mat_<cv::Vec3b>>(m, "cvMat3b", pybind11::buffer_protocol())
        .def(pybind11::init(&npBufferToCvMat3b))
        .def_buffer(&cvMat3bToNpBuffer);
    pybind11::class_<cv::Mat_<float>>(m, "cvMatf", pybind11::buffer_protocol())
        .def(pybind11::init(&npBufferToCvMatF))
        .def_buffer(&cvMatFToNpBuffer);

    // Sandbox pybind11 wrapper
    pybind11::class_<Sandbox>(m, "Sandbox")
        .def(pybind11::init<>())
        .def("init_sandbox",  [](Sandbox &self){
            return self.init();
        })
        .def("load_config", [](Sandbox &self, char* path){
            return self.loadConfig(path);
        })
        .def("capture_frame", [](Sandbox &self){
            return self.captureFrame();
        })
        .def("adjust_projection", [](Sandbox &self, cv::Mat_<cv::Vec3b> frame, cv::Mat_<float> depth) {
            return self.adjustProjection(frame, depth);
        })
        .def("get_color_frame", [](Sandbox &self){
            cv::Mat_<cv::Vec3b> out;
            cv::Mat_<cv::Vec3b> cropped = self.getColorFrame();
            cropped.copyTo(out);
            return out;
            // return giveMatrix();
        })
        .def("get_depth_frame", [](Sandbox &self){
            cv::Mat_<float> out;
            cv::Mat_<float> cropped = self.getDepthFrame();
            cropped.copyTo(out);
            return out;
            // return giveFloatMatrix();
        })
        .def("get_distance_top", [](Sandbox &self){
            return self.getProjection()->getDistanceTopSandbox();
        })
        ;
}