import numpy as np
import cv2
import time
from .wrapper import sandbox_wrapper as sw
# import sandbox_wrapper as sw

WINDOW_NAME='Remote'
ENABLE_WINDOW = True

class Sandbox():
    class SandboxException(Exception):
        ...

    _cppbox=None
    _verbosity = 0
    # running = False
    # refresh_rate = 500 #ms
    # on_frame = None
    # initialized = False

    # display=ENABLE_WINDOW

    def __init__(self, refresh=100, display_enabled=ENABLE_WINDOW) -> None:
        """
        Instantiate the sandbox.

        Arguments:
            refresh : The refresh rate in miliseconds
            display_enabled : If set to false, won't display a window
        """
        self._cppbox = sw.Sandbox()
        self.refresh_rate = refresh
        self.display = display_enabled
        self.running = False
        self.on_frame = None
        self.initialized = False


    def init(self, config_file=None) -> bool:
        """
        Arguments:
            config_file : A path to a sandbox configuration file. If omitted,
            look for a file named "sandbox_conf.yaml" in the current directory.
        """
        if config_file is None:
            config_file = "./sandbox_conf.yaml"
        loaded = self._cppbox.load_config(config_file)
        inited = self._cppbox.init_sandbox()
        self.initialized = not loaded and not inited
        return self.initialized
    
    def start(self, *args, **kwargs) -> None:
        """
        Start the sandbox.
        
        Arguments:
            All arguments are passed down to the potential "on_frame"
            callback function       
        """
        self._main_loop(args, kwargs)

    def stop(self) -> None:
        """
        Stop the sandbox.
        """
        self.running = False

    def capture(self) -> None:
        """
        Order a capture from the camera
        """
        self._cppbox.capture_frame()
        # return True
    
    def get_depth_frame(self) -> np.ndarray:
        """
        Get the depth frame of the sandbox

        Return:
            A 1-channel numpy array of floats
        """
        return np.array(self._cppbox.get_depth_frame(), copy=False)

    def get_color_frame(self) -> np.ndarray:
        """
        Get the color frame of the sandbox

        Return:
            A 3-channel numpy array of unsigned bytes
        """
        return np.array(self._cppbox.get_color_frame(), copy=False)

    def adjust_projection(self, frame, depth) -> np.ndarray:
        """
        Adjust the frame to be correctly projected from the beamer.

        Return:
            the adjusted frame
        """
        frame = self._cppbox.adjust_projection(sw.cvMat3b(frame), sw.cvMatf(depth))
        return np.array(frame)

    def get_distance_top(self) -> float:
        """
        Get the distance from the top of the sandbox.
        
        Return:
            A float (Distance from top, according to calibration)
        """
        return self._cppbox.get_distance_top()

    def _main_loop(self, func_args=(), func_kwargs={}) -> None:
        """
        Main loop. Shouldn't be used directly, call start() instead.

        When called, begin to run the sandbox.
        Perform a capture, get the frames, call a function on it (on_frame),
        adjust the projection and display it using opencv.

        Arguments:
            func_args : A n-tuple of args. Passed to the callback
            func_kwargs : A dictionnary of named arguments. Passed to the callback
        """
        if not self.initialized:
            raise Sandbox.SandboxException("Not initialized. Cannot run.")
        self.running = True
        # Create window
        if self.display:
            cv2.namedWindow(WINDOW_NAME, cv2.WND_PROP_FULLSCREEN)
            cv2.setWindowProperty(WINDOW_NAME, cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN)
        now = time.time() # For debug purposes.
        # Begin main loop.
        while(self.running):
            # Order a capture
            self.capture()
            # Take the current frame (Mandatory ?)
            curr_frame = self.get_color_frame()
            # Take the depth frame of the camera
            curr_depth = self.get_depth_frame()

            # If we have a function binded here, execute it.
            if self.on_frame:
                # Override the current frame
                curr_frame = self.on_frame(curr_depth, curr_frame, *func_args, **func_kwargs)

            # Adjust the projection according to the sandbox parameter
            curr_frame = self.adjust_projection(curr_frame, curr_depth)

            if self.display:
                # Show the result
                cv2.imshow(winname=WINDOW_NAME, mat=curr_frame)
                if cv2.getWindowProperty(WINDOW_NAME,cv2.WND_PROP_VISIBLE) < 1:        
                    break
                if cv2.waitKey(self.refresh_rate) == 27:
                    break
            else:
                time.sleep(self.refresh_rate * 0.001)
            # Compute loop time, if we are verbose.        
            if self._verbosity > 2:
                ti = time.time()
                t_diff = ti - now
                now = ti
                print(f"\tLoop time: {t_diff:.4f} sec\t", end="\r")
            
        # End the loop, destroy everything, output a new line to avoid loosing the last output.
        if self._verbosity > 2: print("\n") 
        if self.display:
            cv2.destroyAllWindows()
        # Set running to false
        self.running = False


class DummySandbox(Sandbox):
    def __init__(self, im, refresh=100, display_enabled=False, noise=True) -> None:
        self.noise = noise
        super().__init__(refresh, display_enabled)
        if type(im) == np.ndarray:
            if im.ndim == 1:
                self.im = im
                self.imc = np.dstack((im, im, im))
            elif im.ndim == 3:
                self.imc = im
                self.im = self.imc[:,:,0]
            else:
                raise Sandbox.SandboxException("Failure.")
        else:
            self.im = cv2.imread(im, flags=cv2.IMREAD_GRAYSCALE)
            self.imc = cv2.imread(im, flags=cv2.IMREAD_COLOR)
        self.factor = 1.0/255.0
    
    def init(self, config_file=None):
        if config_file is None:
            config_file = "./sandbox_conf.yaml"
        loaded = self._cppbox.load_config(config_file)
        inited = self._cppbox.init_sandbox()
        inited = 0 # We don't care if we can't connect to the devices
        # inverted logic, 0 is good. We don't care about inited
        self.initialized = not loaded and not inited
        return self.initialized
    

    def capture(self):
        pass

    def get_depth_frame(self):
        n = np.array(self.im).astype(np.float32)
        n *= self.factor
        return n

    def get_color_frame(self):
        if self.noise:
            mat = np.random.rand(self.imc.shape[0], self.imc.shape[1], self.imc.shape[2])
            mat *= 255
            return mat.astype(np.uint8)
        else: return self.imc

    def adjust_projection(self, frame, depth):
        return frame